import logging
import os
from typing import Callable, Sequence

import numpy as np

NUM_ITERS_MAX = 100
TESTING_MODE = os.getenv("PYTEST_CURRENT_TEST") is not None


def _iter_solver(
    initial_value: np.ndarray,
    objective_value: np.ndarray,
    func_eval: Callable[[np.ndarray | float], float],
    initial_increment: float = 4.0,
    num_iters_max: int = NUM_ITERS_MAX,
    precision: float = 0.01,
) -> tuple[float, int]:
    """Solve by iteration."""
    decreasing = True
    increment = initial_increment
    num_iter = 0
    value_calc = initial_value.copy()
    error = objective_value - func_eval(initial_value)
    while abs(error) > precision and num_iter < num_iters_max:
        iteration_value = func_eval(value_calc)
        error = objective_value - iteration_value
        if abs(error) < precision:
            break
        if error < 0:
            if not decreasing:
                increment /= 2
                decreasing = True
            increment = max(precision / 20, increment)
            value_calc -= increment
        else:
            if decreasing:
                increment /= 2
                decreasing = False
            increment = max(precision / 20, increment)
            value_calc += increment
        num_iter += 1

        if num_iter == num_iters_max:  # pragma: no cover
            raise AssertionError(
                f"No convergence error after {num_iter} iterations! "
                f"Last value: {value_calc}, ∆: {increment}. "
                f"Objective: {objective_value}, iter_value: {iteration_value}"
            )
    return value_calc, num_iter


def solve_curves_with_iteration(
    family_name,
    objective_values: np.ndarray,
    func_init: Callable[[np.ndarray], float],
    func_eval: Callable[[np.ndarray | float], float],
) -> np.ndarray:
    """Run the iteration solver for a list of objective values
    for the three types of curves solved with this method."""
    # family:= checking precision | initial_increment | precision
    families = {
        "ENTHALPHY": (0.01, 0.5, 0.01),
        "CONSTANT VOLUME": (0.0005, 1, 0.00025),
    }
    # "CONSTANT VOLUME": (0.0005, 1, 0.00000025, 0.0025, 0.75, 0.00000025),
    if family_name not in families.keys():  # pragma: no cover
        raise AssertionError(
            f"Need a valid family of curves: {list(families.keys())}"
        )

    precision_comp, initial_increment, precision = families[family_name]
    calc_points: list[float] = []
    for objective in objective_values:
        try:
            calc_p, num_iter = _iter_solver(
                np.array(func_init(objective)),
                np.array(objective),
                func_eval=func_eval,
                initial_increment=initial_increment,
                precision=precision,
            )
        except AssertionError as exc:  # pragma: no cover
            logging.error(f"{family_name} CONVERGENCE ERROR: {exc}")
            if TESTING_MODE:
                raise exc
            else:
                return np.array(calc_points)

        if TESTING_MODE and (
            abs(objective - func_eval(calc_p)) > precision_comp
        ):  # pragma: no cover
            msg = (
                f"{family_name} BAD RESULT[#{num_iter}] "
                f"(E={abs(objective - func_eval(calc_p)):.5f}): "
                f"objective: {objective:.5f}, calc_p: {calc_p:.5f}, "
                f"EVAL: {func_eval(calc_p):.5f}"
            )
            logging.error(msg)
            raise AssertionError(msg)
        calc_points.append(calc_p)
    return np.array(calc_points)


def mod_color(color: Sequence[float], modification: float) -> list[float]:
    """Modify color with an alpha value or a darken/lighten percentage."""
    if abs(modification) < 0.999:  # is alpha level
        color = list(color[:3]) + [modification]
    else:
        color = [
            max(0.0, min(1.0, c * (1 + modification / 100))) for c in color
        ]
    return color



from matplotlib.colors import to_rgba
import numpy as np


def parse_color(raw_color) -> list[float]:
    """Pydantic validator for 'color' fields, stored as float values."""
    if isinstance(raw_color, str):
        return list(to_rgba(raw_color))
    return list(raw_color)


def reduce_field_abrs(values):
    """Pydantic validator for abbreviation of matplotlib style fields."""
    if "c" in values:
        values["color"] = parse_color(values.pop("c"))
    if "ls" in values:
        values["linestyle"] = values.pop("ls")
    if "lw" in values:
        values["linewidth"] = values.pop("lw")
    return values


def check_connector_and_areas_by_point_name(values):
    """Pydantic validator for ChartAnnots, removing unknown points."""
    keys_1 = set(values.get("points", {}).keys())
    keys_2 = set(values.get("series", {}).keys())
    valid_keys = keys_1 | keys_2
    values["connectors"] = [
        cn
        for cn in values["connectors"]
        if cn.start in valid_keys and cn.end in valid_keys
    ]
    valid_areas = []
    for chart_area in values["areas"]:
        point_names = [
            name for name in chart_area.point_names if name in valid_keys
        ]
        if len(point_names) >= 3:
            chart_area.point_names = point_names
            valid_areas.append(chart_area)

    values["areas"] = valid_areas
    return values


def parse_curve_arrays(values):
    """Pydantic validator to convert values into np.arrays."""
    if "x_data" in values and not isinstance(values["x_data"], np.ndarray):
        values["x_data"] = np.array(values["x_data"])
    if "y_data" in values and not isinstance(values["y_data"], np.ndarray):
        values["y_data"] = np.array(values["y_data"])
    shape = values["x_data"].shape[0], values["y_data"].shape[0]
    if shape[0] == 0 or shape[1] == 0 or shape[0] != shape[1]:
        raise ValueError(f"Invalid shape: {shape}")
    return values



"""A library to make psychrometric charts and overlay information in them."""
from psychrochart.chart import PsychroChart
from psychrochart.models.annots import ChartAnnots, ChartZones
from psychrochart.models.config import ChartConfig
from psychrochart.models.curves import PsychroCurve, PsychroCurves
from psychrochart.models.parsers import load_config, load_zones
from psychrochart.models.styles import (
    CurveStyle,
    LabelStyle,
    TickStyle,
    ZoneStyle,
)

__all__ = [
    "ChartAnnots",
    "ChartConfig",
    "ChartZones",
    "load_config",
    "load_zones",
    "PsychroChart",
    "PsychroCurve",
    "PsychroCurves",
    "CurveStyle",
    "LabelStyle",
    "TickStyle",
    "ZoneStyle",
]



"""A library to make psychrometric charts and overlay information in them."""
import logging
from typing import Sequence

import numpy as np
from psychrolib import (
    GetHumRatioFromTWetBulb,
    GetHumRatioFromVapPres,
    GetMoistAirEnthalpy,
    GetMoistAirVolume,
    GetRelHumFromHumRatio,
    GetSatVapPres,
    GetTDewPointFromVapPres,
    GetTDryBulbFromEnthalpyAndHumRatio,
    GetTDryBulbFromMoistAirVolumeAndHumRatio,
    GetTWetBulbFromHumRatio,
    GetVapPresFromHumRatio,
    isIP,
)
from scipy.interpolate import interp1d

from psychrochart.chart_entities import random_internal_value
from psychrochart.models.annots import ChartZone
from psychrochart.models.curves import PsychroCurve, PsychroCurves
from psychrochart.models.styles import CurveStyle
from psychrochart.util import solve_curves_with_iteration

f_vec_hum_ratio_from_vap_press = np.vectorize(GetHumRatioFromVapPres)
f_vec_moist_air_enthalpy = np.vectorize(GetMoistAirEnthalpy)
f_vec_moist_air_volume = np.vectorize(GetMoistAirVolume)
f_vec_dew_point_from_vap_press = np.vectorize(GetTDewPointFromVapPres)
f_vec_dry_temp_from_enthalpy = np.vectorize(GetTDryBulbFromEnthalpyAndHumRatio)
f_vec_dry_temp_from_spec_vol = np.vectorize(
    GetTDryBulbFromMoistAirVolumeAndHumRatio
)
f_vec_sat_press = np.vectorize(GetSatVapPres)
f_vec_vap_press_from_hum_ratio = np.vectorize(GetVapPresFromHumRatio)


def _factor_out_w() -> float:
    """
    Conversion factor from internal units to plot units for humidity ratio.

    In SI, w is internally in kg(w)/kg(da), but for plots we use g(w)/kg(da).
    """
    return 7000.0 if isIP() else 1000.0


def _factor_out_h() -> float:
    """
    Conversion factor from internal units to plot units for enthalpy.

    In SI, h is internally in J/kg, but for plots we use kJ/kg.
    """
    return 1.0 if isIP() else 1000.0


def _make_enthalpy_label(enthalpy: float) -> str:
    unit = "$Btu/lb_{da}$" if isIP() else "$kJ/kg_{da}$"
    return f"{enthalpy:g} {unit}"


def _make_temp_label(temperature: float) -> str:
    unit = "°F" if isIP() else "°C"
    return f"{temperature:g} {unit}"


def _make_vol_label(specific_vol: float) -> str:
    unit = "$ft³/lb_{da}$" if isIP() else "$m³/kg_{da}$"
    return f"{specific_vol:g} {unit}"


def _get_humid_ratio_in_saturation(
    dry_temps: np.ndarray, pressure: float
) -> np.array:
    sat_p = f_vec_sat_press(dry_temps)
    return _factor_out_w() * f_vec_hum_ratio_from_vap_press(sat_p, pressure)


def get_rh_max_min_in_limits(
    dbt_min: float,
    dbt_max: float,
    w_humidity_ratio_min: float,
    w_humidity_ratio_max: float,
    pressure: float,
) -> tuple[float, float]:
    """Get max range for constant rel. humidity lines inside chart limits."""
    rh_min = GetRelHumFromHumRatio(
        dbt_max, w_humidity_ratio_min / _factor_out_w(), pressure
    )
    rh_max = GetRelHumFromHumRatio(
        dbt_min, w_humidity_ratio_max / _factor_out_w(), pressure
    )
    return max(rh_min * 100.0, 0), min(rh_max * 100.0, 100)


def gen_points_in_constant_relative_humidity(
    dry_temps: Sequence[float],
    rh_percentage: float | Sequence[float],
    pressure: float,
) -> np.array:
    """Generate a curve (numpy array) of constant humidity ratio."""
    return _factor_out_w() * f_vec_hum_ratio_from_vap_press(
        f_vec_sat_press(dry_temps)
        * np.array(rh_percentage).clip(0, 100)
        / 100.0,
        pressure,
    )


def make_constant_relative_humidity_lines(
    dbt_min: float,
    dbt_max: float,
    temp_step: float,
    pressure: float,
    rh_perc_values: list[int],
    rh_label_values: list[int],
    style: CurveStyle,
    label_loc: float,
    family_label: str | None,
) -> PsychroCurves:
    """Generate curves of constant relative humidity for the chart."""
    rh_values = sorted(rh for rh in rh_perc_values if 0 <= rh <= 100)
    temps_ct_rh = np.arange(dbt_min, dbt_max + temp_step, temp_step)
    curves_ct_rh = [
        gen_points_in_constant_relative_humidity(temps_ct_rh, rh, pressure)
        for rh in rh_values
    ]
    return PsychroCurves(
        curves=[
            PsychroCurve(
                x_data=temps_ct_rh,
                y_data=curve_ct_rh,
                style=style,
                type_curve="constant_rh_data",
                label_loc=label_loc,
                label=f"RH {rh:g} %" if rh in rh_label_values else None,
                internal_value=float(rh),
            )
            for rh, curve_ct_rh in zip(rh_values, curves_ct_rh)
        ],
        family_label=family_label,
    )


def make_constant_dry_bulb_v_line(
    w_humidity_ratio_min: float,
    temp: float,
    pressure: float,
    style: CurveStyle,
    type_curve: str | None = None,
    reverse: bool = False,
) -> PsychroCurve:
    """Generate vertical line (constant dry bulb temp) up to saturation."""
    w_max = _factor_out_w() * GetHumRatioFromVapPres(
        GetSatVapPres(temp), pressure
    )
    if reverse:
        path_y = [w_max, w_humidity_ratio_min]
    else:
        path_y = [w_humidity_ratio_min, w_max]
    return PsychroCurve(
        x_data=np.array([temp, temp]),
        y_data=np.array(path_y),
        style=style,
        type_curve=type_curve,
        internal_value=temp,
    )


def make_constant_dry_bulb_v_lines(
    w_humidity_ratio_min: float,
    pressure: float,
    temps_vl: np.ndarray,
    style: CurveStyle,
    family_label: str | None,
) -> PsychroCurves:
    """Generate curves of constant dry bulb temperature (vertical)."""
    w_max_vec = _get_humid_ratio_in_saturation(temps_vl, pressure)
    return PsychroCurves(
        curves=[
            PsychroCurve(
                x_data=np.array([temp, temp]),
                y_data=np.array([w_humidity_ratio_min, w_max]),
                style=style,
                type_curve="constant_dry_temp_data",
                internal_value=temp,
            )
            for temp, w_max in zip(temps_vl, w_max_vec)
        ],
        family_label=family_label,
    )


def make_constant_humidity_ratio_h_lines(
    dbt_max: float,
    pressure: float,
    ws_hl: np.ndarray,
    style: CurveStyle,
    family_label: str | None,
) -> PsychroCurves:
    """Generate curves of constant absolute humidity (horizontal)."""
    arr_hum_ratios = np.array(ws_hl) / _factor_out_w()
    dew_points = f_vec_dew_point_from_vap_press(
        dbt_max, f_vec_vap_press_from_hum_ratio(arr_hum_ratios, pressure)
    )
    return PsychroCurves(
        curves=[
            PsychroCurve(
                x_data=np.array([t_dp, dbt_max]),
                y_data=np.array([w, w]),
                style=style,
                type_curve="constant_humidity_data",
                internal_value=w,
            )
            for w, t_dp in zip(ws_hl, dew_points)
        ],
        family_label=family_label,
    )


def make_saturation_line(
    dbt_min: float,
    dbt_max: float,
    temp_step: float,
    pressure: float,
    style: CurveStyle,
) -> PsychroCurves:
    """Generate line of saturation for the psychrochart."""
    temps_sat_line = np.arange(dbt_min, dbt_max + temp_step, temp_step)
    w_sat = gen_points_in_constant_relative_humidity(
        temps_sat_line, 100, pressure
    )
    sat_c = PsychroCurve(
        x_data=temps_sat_line,
        y_data=w_sat,
        style=style,
        type_curve="saturation",
        internal_value=100.0,
    )
    return PsychroCurves(curves=[sat_c])


def make_constant_enthalpy_lines(
    w_humidity_ratio_min: float,
    pressure: float,
    enthalpy_values: np.ndarray,
    h_label_values: list[float],
    style: CurveStyle,
    label_loc: float,
    family_label: str | None,
    saturation_curve: PsychroCurve,
    dbt_min_seen: float | None = None,
) -> PsychroCurves | None:
    """Generate curves of constant enthalpy for the chart."""
    h_in_sat = (
        f_vec_moist_air_enthalpy(
            saturation_curve.x_data, saturation_curve.y_data / _factor_out_w()
        )
        / _factor_out_h()
    )
    t_sat_interpolator = interp1d(
        h_in_sat,
        saturation_curve.x_data,
        fill_value="extrapolate",
        assume_sorted=True,
    )
    h_min = (
        GetMoistAirEnthalpy(
            dbt_min_seen or saturation_curve.x_data[0],
            w_humidity_ratio_min / _factor_out_w(),
        )
        / _factor_out_h()
    )
    h_max = h_in_sat[-1]
    h_objective = np.array([h for h in enthalpy_values if h_min < h < h_max])
    if not h_objective.shape[0]:
        logging.warning(
            "All %d enthalpy curves are outside limits"
            "(%g->%g not inside [%g, %g])",
            len(enthalpy_values),
            enthalpy_values[0],
            enthalpy_values[-1],
            h_min,
            h_max,
        )
        return None

    temps_max_constant_h = f_vec_dry_temp_from_enthalpy(
        h_objective * _factor_out_h(),
        w_humidity_ratio_min / _factor_out_w(),
    )
    t_sat_points = solve_curves_with_iteration(
        "ENTHALPHY",
        h_objective,
        lambda *x: t_sat_interpolator(x[0]),
        lambda x: GetMoistAirEnthalpy(
            x, GetHumRatioFromVapPres(GetSatVapPres(x), pressure)
        )
        / _factor_out_h(),
    )
    w_in_sat = _get_humid_ratio_in_saturation(t_sat_points, pressure)

    return PsychroCurves(
        curves=[
            PsychroCurve(
                x_data=np.array([t_sat, t_max]),
                y_data=np.array([w_sat, w_humidity_ratio_min]),
                style=style,
                type_curve="constant_h_data",
                label_loc=label_loc,
                label=(
                    _make_enthalpy_label(h)
                    if round(h, 3) in h_label_values
                    else None
                ),
                internal_value=round(h, 3),
            )
            for t_sat, w_sat, t_max, h in zip(
                t_sat_points, w_in_sat, temps_max_constant_h, h_objective
            )
        ],
        family_label=family_label,
    )


def make_constant_specific_volume_lines(
    w_humidity_ratio_min: float,
    pressure: float,
    vol_values: np.ndarray,
    v_label_values: list[float],
    style: CurveStyle,
    label_loc: float,
    family_label: str | None,
    saturation_curve: PsychroCurve,
    dbt_min_seen: float | None = None,
) -> PsychroCurves | None:
    """Generate curves of constant specific volume for the chart."""
    v_in_sat = f_vec_moist_air_volume(
        saturation_curve.x_data,
        saturation_curve.y_data / _factor_out_w(),
        pressure,
    )
    v_min = GetMoistAirVolume(
        dbt_min_seen or saturation_curve.x_data[0],
        w_humidity_ratio_min / _factor_out_w(),
        pressure,
    )
    v_max = v_in_sat[-1]
    valid_objectives = [h for h in vol_values if v_min < h < v_max]
    if not valid_objectives:
        logging.warning(
            "All %d constant-volume curves are outside limits "
            "(%g->%g not inside [%g, %g])",
            len(vol_values),
            vol_values[0],
            vol_values[-1],
            v_min,
            v_max,
        )
        return None

    v_objective = np.array(valid_objectives)
    temps_max_constant_v = f_vec_dry_temp_from_spec_vol(
        np.array(v_objective), w_humidity_ratio_min / _factor_out_w(), pressure
    )
    t_sat_interpolator = interp1d(
        v_in_sat,
        saturation_curve.x_data,
        fill_value="extrapolate",
        assume_sorted=True,
    )
    t_sat_points = solve_curves_with_iteration(
        "CONSTANT VOLUME",
        v_objective,
        lambda *x: t_sat_interpolator(x[0]),
        lambda x: GetMoistAirVolume(
            x, GetHumRatioFromVapPres(GetSatVapPres(x), pressure), pressure
        ),
    )
    w_in_sat = _get_humid_ratio_in_saturation(t_sat_points, pressure)

    return PsychroCurves(
        curves=[
            PsychroCurve(
                x_data=np.array([t_sat, t_max]),
                y_data=np.array([w_sat, w_humidity_ratio_min]),
                style=style,
                type_curve="constant_v_data",
                label_loc=label_loc,
                label=(
                    _make_vol_label(vol)
                    if round(vol, 3) in v_label_values
                    else None
                ),
                internal_value=round(vol, 3),
            )
            for t_sat, w_sat, t_max, vol in zip(
                t_sat_points, w_in_sat, temps_max_constant_v, v_objective
            )
        ],
        family_label=family_label,
    )


def make_constant_wet_bulb_temperature_lines(
    dry_bulb_temp_min: float,
    dry_bulb_temp_max: float,
    w_humidity_ratio_min: float,
    w_humidity_ratio_max: float,
    pressure: float,
    wbt_values: np.ndarray,
    wbt_label_values: list[float],
    style: CurveStyle,
    label_loc: float,
    family_label: str | None,
) -> PsychroCurves | None:
    """Generate curves of constant wet bulb temperature for the chart."""
    wt_min = GetTWetBulbFromHumRatio(
        dry_bulb_temp_min, w_humidity_ratio_min / _factor_out_w(), pressure
    )
    if -0.75 < wt_min < 0.0:  # slope change zone
        wt_min = 0
    wt_bottom_right = GetTWetBulbFromHumRatio(dry_bulb_temp_max, 0, pressure)
    wt_top_right = GetTWetBulbFromHumRatio(
        dry_bulb_temp_max,
        min(
            w_humidity_ratio_max / _factor_out_w(),
            GetHumRatioFromVapPres(GetSatVapPres(dry_bulb_temp_max), pressure),
        ),
        pressure,
    )
    wbt_objective = np.array(
        [wt for wt in wbt_values if wt_min < wt < wt_top_right]
    )
    if wbt_objective.shape[0] == 0:
        logging.warning(
            "All %d wetbulb-temp curves are outside limits "
            "(%g->%g not inside [%g, %g])",
            len(wbt_values),
            wbt_values[0],
            wbt_values[-1],
            wt_min,
            dry_bulb_temp_max,
        )
        return None

    w_max_constant_wbt = f_vec_hum_ratio_from_vap_press(
        f_vec_sat_press(wbt_objective), pressure
    )
    curves = []
    for wbt, w_max in zip(wbt_objective, w_max_constant_wbt):
        dbt_objective = dry_bulb_temp_max
        w_left = _factor_out_w() * w_max
        if wbt >= wt_bottom_right:
            # on vertical y-axis
            w_right = _factor_out_w() * GetHumRatioFromTWetBulb(
                dbt_objective, wbt, pressure
            )
        else:
            # on horizontal x-axis
            dbt_objective = dry_bulb_temp_max
            w_right = _factor_out_w() * GetHumRatioFromTWetBulb(
                dbt_objective, wbt, pressure
            )
            while w_right <= 0.01:
                dbt_objective -= 0.5 * (dbt_objective - wbt)
                w_right = _factor_out_w() * GetHumRatioFromTWetBulb(
                    dbt_objective, wbt, pressure
                )
                if w_right > 0.01:
                    # extend curve to the bottom axis
                    slope = (dbt_objective - wbt) / (w_right - w_left)
                    dbt_objective = wbt - slope * w_left
                    w_right = 0.0
                    break

        c = PsychroCurve(
            x_data=np.array([wbt, dbt_objective]),
            y_data=np.array([w_left, w_right]),
            style=style,
            type_curve="constant_wbt_data",
            label_loc=label_loc,
            label=(_make_temp_label(wbt) if wbt in wbt_label_values else None),
            internal_value=wbt,
        )
        curves.append(c)

    return PsychroCurves(curves=curves, family_label=family_label)


def make_zone_curve(
    zone_conf: ChartZone, increment: float, pressure: float
) -> PsychroCurve:
    """Generate plot-points for zone."""
    # todo better id for overlay zones if no label
    zone_value = random_internal_value() if zone_conf.label is None else None
    if zone_conf.zone_type == "xy-points":
        # expect points in plot coordinates!
        return PsychroCurve(
            x_data=np.array(zone_conf.points_x),
            y_data=np.array(zone_conf.points_y),
            style=zone_conf.style,
            type_curve="xy-points",
            label=zone_conf.label,
            internal_value=zone_value,
        )

    assert zone_conf.zone_type == "dbt-rh"
    # points for zone between constant dry bulb temps and RH
    t_min = zone_conf.points_x[0]
    t_max = zone_conf.points_x[-1]
    rh_min = zone_conf.points_y[0]
    rh_max = zone_conf.points_y[-1]
    assert rh_min >= 0.0 and rh_max <= 100.0
    assert t_min < t_max

    temps = np.arange(t_min, t_max + increment, increment)
    curve_rh_up = gen_points_in_constant_relative_humidity(
        temps, rh_max, pressure
    )
    curve_rh_down = gen_points_in_constant_relative_humidity(
        temps, rh_min, pressure
    )
    abs_humid: list[float] = (
        list(curve_rh_up) + list(curve_rh_down)[::-1] + [curve_rh_up[0]]
    )
    temps_zone: list[float] = list(temps) + list(temps)[::-1] + [temps[0]]
    return PsychroCurve(
        x_data=np.array(temps_zone),
        y_data=np.array(abs_humid),
        style=zone_conf.style,
        type_curve="dbt-rh",
        label=zone_conf.label,
        internal_value=zone_value,
    )



from typing import Any, Literal

import numpy as np
from pydantic import BaseModel, Field, root_validator

from psychrochart.models.styles import CurveStyle, ZoneStyle
from psychrochart.models.validators import (
    check_connector_and_areas_by_point_name,
    parse_curve_arrays,
)

ConvexGroupTuple = tuple[list[str], dict[str, Any], dict[str, Any]]


class ChartPoint(BaseModel):
    """Pydantic model for single point annotation."""

    xy: tuple[float | int, float | int]
    label: str | None = Field(default=None)
    style: dict[str, Any] = Field(default_factory=dict)


class ChartSeries(BaseModel):
    """Pydantic model for data-series point array annotation."""

    # TODO fusion with PsychroCurve, + pandas ready
    x_data: np.ndarray
    y_data: np.ndarray
    style: dict[str, Any] = Field(default_factory=dict)
    label: str | None = None

    class Config:
        arbitrary_types_allowed = True
        json_encoders = {np.ndarray: lambda x: x.tolist()}

    @root_validator(pre=True)
    def _parse_curve_data(cls, values):
        return parse_curve_arrays(values)


class ChartLine(BaseModel):
    """Pydantic model for a connector between 2 named points."""

    start: str
    end: str
    label: str | None = Field(default=None)
    style: CurveStyle = Field(default_factory=CurveStyle)
    outline_marker_width: int | None = Field(default=None)


class ChartArea(BaseModel):
    """Pydantic model for convex area delimited by a list of named points."""

    point_names: list[str] = Field(min_items=3)
    line_style: dict[str, Any] = Field(default_factory=dict)
    fill_style: dict[str, Any] = Field(default_factory=dict)
    # label: str | None = None

    @classmethod
    def from_tuple(cls, item: ConvexGroupTuple):
        """
        Parse area definition from old tuples.

        The syntax to add groups of given points (with more than 3 points)
        to plot a styled convex hull area was:
        ```
        interior_zones = [
            # Zone 1:
            ([point_1_name, point_2_name, point_3_name, ...],
             {"color": 'darkgreen', "lw": 0, ...},
             {"color": 'darkgreen', "lw": 0, ...}),

            # Zone 2:
            ([point_7_name, point_8_name, point_9_name, ...],
             {"color": 'darkorange', "lw": 0, ...},
             {"color": 'darkorange', "lw": 0, ...}),

            # ...
        ]
        """
        return cls.validate(
            dict(zip(["point_names", "line_style", "fill_style"], item))
        )


class ChartAnnots(BaseModel):
    """Container model for all chart annots based on overlayed data points."""

    points: dict[str, ChartPoint] = Field(default_factory=dict)
    series: dict[str, ChartSeries] = Field(default_factory=dict)
    connectors: list[ChartLine] = Field(default_factory=list)
    areas: list[ChartArea] = Field(default_factory=list)
    use_scatter: bool = Field(default=False)

    @root_validator
    def _validate_used_points(cls, values):
        return check_connector_and_areas_by_point_name(values)

    def get_point_by_name(self, key: str) -> tuple[float, float]:
        """Access coords tuple for named point (or 1st point of series)."""
        if key in self.points:
            return self.points[key].xy
        else:
            assert key in self.series
            return self.series[key].x_data[0], self.series[key].y_data[0]


class ChartZone(BaseModel):
    """Pydantic model for styled fixed areas on the psychrochart."""

    label: str | None
    zone_type: Literal["dbt-rh", "xy-points"]
    points_x: list[float]
    points_y: list[float]
    style: ZoneStyle


class ChartZones(BaseModel):
    """Container model for a list of ChartZone items."""

    zones: list[ChartZone]


# default fixed areas for winter/summer comfort zones
DEFAULT_ZONES = ChartZones(
    zones=[
        ChartZone(
            label="Summer",
            zone_type="dbt-rh",
            points_x=[23, 25],
            points_y=[45, 60],
            style=ZoneStyle(
                edgecolor=[1.0, 0.749, 0.0, 0.8],
                facecolor=[1.0, 0.749, 0.0, 0.5],
            ),
        ),
        ChartZone(
            label="Winter",
            zone_type="dbt-rh",
            points_x=[21, 23],
            points_y=[40, 50],
            style=ZoneStyle(
                edgecolor=[0.498, 0.624, 0.8],
                facecolor=[0.498, 0.624, 1.0, 0.5],
            ),
        ),
    ]
)



"""A library to make psychrometric charts and overlay information in them."""
import logging
from math import atan2, degrees
from typing import Any, AnyStr

from matplotlib import patches
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.path import Path
from matplotlib.text import Annotation
import numpy as np
from scipy.spatial import ConvexHull, QhullError

from psychrochart.models.annots import ChartAnnots
from psychrochart.models.config import ChartConfig
from psychrochart.models.curves import (
    PsychroChartModel,
    PsychroCurve,
    PsychroCurves,
)
from psychrochart.models.styles import ZoneStyle
from psychrochart.util import mod_color


def _annotate_label(
    ax: Axes,
    label: AnyStr,
    text_x: float,
    text_y: float,
    rotation: float,
    text_style: dict[str, Any],
) -> Annotation:
    if abs(rotation) > 0:
        text_loc = np.array((text_x, text_y))
        text_style["rotation"] = ax.transData.transform_angles(
            np.array((rotation,)), text_loc.reshape((1, 2))
        )[0]
        text_style["rotation_mode"] = "anchor"
    return ax.annotate(label, (text_x, text_y), **text_style)


def add_label_to_curve(
    curve: PsychroCurve,
    ax: Axes,
    text_label: str | None = None,
    va: str | None = None,
    ha: str | None = None,
    loc: float | None = None,
    **params,
) -> Annotation:
    """Annotate the curve with its label."""
    num_samples = len(curve.x_data)
    assert num_samples > 1
    text_style = {"va": "bottom", "ha": "left", "color": [0.0, 0.0, 0.0]}
    loc_f: float = curve.label_loc if loc is None else loc
    label: str = (
        (curve.label if curve.label is not None else "")
        if text_label is None
        else text_label
    )

    def _tilt_params(x_data, y_data, idx_0, idx_f):
        delta_x = x_data[idx_f] - curve.x_data[idx_0]
        delta_y = y_data[idx_f] - curve.y_data[idx_0]
        rotation_deg = degrees(atan2(delta_y, delta_x))
        if delta_x == 0:
            tilt_curve = 1e12
        else:
            tilt_curve = delta_y / delta_x
        return rotation_deg, tilt_curve

    if num_samples == 2:
        xmin, xmax = ax.get_xlim()
        rotation, tilt = _tilt_params(curve.x_data, curve.y_data, 0, 1)
        if abs(rotation) == 90:
            text_x = curve.x_data[0]
            text_y = curve.y_data[0] + loc_f * (
                curve.y_data[1] - curve.y_data[0]
            )
        elif loc_f == 1.0:
            if curve.x_data[1] > xmax:
                text_x = xmax
                text_y = curve.y_data[0] + tilt * (xmax - curve.x_data[0])
            else:
                text_x, text_y = curve.x_data[1], curve.y_data[1]
            label += "    "
            text_style["ha"] = "right"
        else:
            text_x = curve.x_data[0] + loc_f * (xmax - xmin)
            if text_x < xmin:
                text_x = xmin + loc_f * (xmax - xmin)
            text_y = curve.y_data[0] + tilt * (text_x - curve.x_data[0])
    else:
        idx = min(num_samples - 2, int(num_samples * loc_f))
        rotation, tilt = _tilt_params(curve.x_data, curve.y_data, idx, idx + 1)
        text_x, text_y = curve.x_data[idx], curve.y_data[idx]
        text_style["ha"] = "center"

    text_style["color"] = mod_color(curve.style.color, -25)
    if ha is not None:
        text_style["ha"] = ha
    if va is not None:
        text_style["va"] = va
    if params:
        text_style.update(params)

    return _annotate_label(ax, label, text_x, text_y, rotation, text_style)


def plot_curve(
    curve: PsychroCurve, ax: Axes, label_prefix: str | None = None
) -> list[Artist]:
    """Plot the curve, if it's between chart limits."""
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    if (
        curve.x_data is None
        or curve.y_data is None
        or max(curve.y_data) < ymin
        or max(curve.x_data) < xmin
        or min(curve.y_data) > ymax
        or min(curve.x_data) > xmax
    ):
        logging.info(
            "%s (label:%s) not between limits ([%.2g, %.2g, %.2g, %.2g]) "
            "-> x:%s, y:%s",
            curve.type_curve,
            curve.label or "unnamed",
            xmin,
            xmax,
            ymin,
            ymax,
            curve.x_data,
            curve.y_data,
        )
        return []

    artists = []
    if isinstance(curve.style, ZoneStyle):
        assert len(curve.y_data) > 2
        verts = list(zip(curve.x_data, curve.y_data))
        codes = (
            [Path.MOVETO]
            + [Path.LINETO] * (len(curve.y_data) - 2)
            + [Path.CLOSEPOLY]
        )
        path = Path(verts, codes)
        patch = patches.PathPatch(path, **curve.style.dict())
        ax.add_patch(patch)
        artists.append(patch)

        if curve.label is not None:
            bbox_p = path.get_extents()
            text_x = 0.5 * (bbox_p.x0 + bbox_p.x1)
            text_y = 0.5 * (bbox_p.y0 + bbox_p.y1)
            style_params = {
                "ha": "center",
                "va": "center",
                "backgroundcolor": [1, 1, 1, 0.4],
            }
            assert isinstance(curve.style, ZoneStyle)
            style_params["color"] = mod_color(curve.style.edgecolor, -25)
            artist_label = _annotate_label(
                ax, curve.label, text_x, text_y, 0, style_params
            )
            artists.append(artist_label)
    else:
        artist_line = ax.plot(curve.x_data, curve.y_data, **curve.style.dict())
        artists.append(artist_line)
        if curve.label is not None:
            artists.append(add_label_to_curve(curve, ax))

    return artists


def plot_curves_family(family: PsychroCurves | None, ax: Axes) -> list[Artist]:
    """Plot all curves in the family."""
    artists: list[Artist] = []
    if family is None:
        return artists

    [
        plot_curve(curve, ax, label_prefix=family.family_label)
        for curve in family.curves
    ]
    # Curves family labelling
    if family.curves and family.family_label is not None:
        artist_fam_label = ax.plot(
            [-1],
            [-1],
            label=family.family_label,
            marker="D",
            markersize=10,
            **family.curves[0].style.dict(),
        )
        artists.append(artist_fam_label)

    # return [
    #     art for art in artist_curves + artist_labels if art is not None
    # ]
    # TODO collect artists from plot_curve
    return []


def _apply_spines_style(axes, style, location="right") -> None:
    for key in style:
        try:
            getattr(axes.spines[location], f"set_{key}")(style[key])
        except Exception as exc:  # pragma: no cover
            logging.error(
                f"Error trying to apply spines attrs: {exc}. "
                f"({dir(axes.spines[location])})"
            )


def apply_axis_styling(config: ChartConfig, ax: Axes) -> None:
    """Setup matplotlib Axes object for the chart."""
    ax.yaxis.tick_right()
    ax.yaxis.set_label_position("right")
    ax.set_xlim(config.dbt_min, config.dbt_max)
    ax.set_ylim(config.w_min, config.w_max)
    ax.grid(False, which="both")
    # Apply axis styles
    if config.figure.x_label is not None:
        style_axis = config.figure.x_axis_labels.dict()
        style_axis["fontsize"] *= 1.2
        ax.set_xlabel(config.figure.x_label, **style_axis)
    if config.figure.y_label is not None:
        style_axis = config.figure.y_axis_labels.dict()
        style_axis["fontsize"] *= 1.2
        ax.set_ylabel(config.figure.y_label, **style_axis)
    if config.figure.title is not None:
        ax.set_title(
            config.figure.title,
            fontsize=config.figure.fontsize * 1.5,
            fontweight="bold",
        )

    _apply_spines_style(ax, config.figure.y_axis.dict(), location="right")
    _apply_spines_style(ax, config.figure.x_axis.dict(), location="bottom")
    if config.figure.partial_axis:  # Hide left and top axis
        ax.spines["left"].set_visible(False)
        ax.spines["top"].set_visible(False)
    else:
        _apply_spines_style(ax, config.figure.y_axis.dict(), location="left")
        _apply_spines_style(ax, config.figure.x_axis.dict(), location="top")

    if config.figure.x_axis_ticks is not None:
        ax.tick_params(axis="x", **config.figure.x_axis_ticks.dict())
    if config.figure.y_axis_ticks is not None:
        ax.tick_params(axis="y", **config.figure.y_axis_ticks.dict())

    # set tick labels in main axes
    if config.chart_params.with_constant_dry_temp:
        step_label = config.chart_params.constant_temp_label_step
        if step_label > 0:  # Explicit xticks
            ticks = np.arange(
                config.dbt_min, config.dbt_max + step_label / 10, step_label
            )
            if not config.chart_params.constant_temp_label_include_limits:
                ticks = [
                    t
                    for t in ticks
                    if t not in [config.dbt_min, config.dbt_max]
                ]
            ax.set_xticks(ticks)
            ax.set_xticklabels(
                [f"{t:g}" for t in ticks], **config.figure.x_axis_labels.dict()
            )
    else:
        ax.set_xticks([])

    if config.chart_params.with_constant_humidity:
        step_label = config.chart_params.constant_humid_label_step
        if step_label > 0:  # Explicit xticks
            ticks = np.arange(
                config.w_min, config.w_max + step_label / 10, step_label
            )
            if not config.chart_params.constant_humid_label_include_limits:
                ticks = [
                    t for t in ticks if t not in [config.w_min, config.w_max]
                ]
            ax.set_yticks(ticks)
            ax.set_yticklabels(
                [f"{t:g}" for t in ticks], **config.figure.y_axis_labels.dict()
            )
    else:
        ax.set_yticks([])


def plot_chart(chart: PsychroChartModel, ax: Axes) -> Axes:
    """Plot the psychrochart curves on given Axes."""
    # Plot curves:
    plot_curves_family(chart.constant_dry_temp_data, ax)
    plot_curves_family(chart.constant_humidity_data, ax)
    plot_curves_family(chart.constant_h_data, ax)
    plot_curves_family(chart.constant_v_data, ax)
    plot_curves_family(chart.constant_rh_data, ax)
    plot_curves_family(chart.constant_wbt_data, ax)
    plot_curves_family(chart.saturation, ax)

    # Plot zones:
    for zone in chart.zones:
        plot_curves_family(zone, ax)
    return ax


def plot_annots_dbt_rh(
    ax: Axes, annots: ChartAnnots
) -> list[Artist | list[Artist]]:
    """Plot chat annotations in given matplotlib Axes, return `Artist` objs."""
    _handlers_annotations = []
    for d_con in annots.connectors:
        x_start, y_start = annots.get_point_by_name(d_con.start)
        x_end, y_end = annots.get_point_by_name(d_con.end)
        x_line = [x_start, x_end]
        y_line = [y_start, y_end]
        _handlers_annotations.append(
            ax.plot(
                x_line,
                y_line,
                label=d_con.label,
                dash_capstyle="round",
                **d_con.style.dict(),
            )
        )
        if d_con.outline_marker_width:
            _handlers_annotations.append(
                ax.plot(
                    x_line,
                    y_line,
                    color=[*d_con.style.color[:3], 0.15],
                    lw=d_con.outline_marker_width,
                    solid_capstyle="round",
                )
            )

    forbidden = set()
    if annots.use_scatter:
        f_plot = ax.scatter
        forbidden.add("markersize")
    else:
        f_plot = ax.plot
    for series in annots.series.values():
        style = {k: v for k, v in series.style.items() if k not in forbidden}
        _handlers_annotations.append(
            f_plot(series.x_data, series.y_data, label=series.label, **style)
        )

    for point in annots.points.values():
        style = {k: v for k, v in point.style.items() if k not in forbidden}
        _handlers_annotations.append(
            f_plot(point.xy[0], point.xy[1], label=point.label, **style)
        )

    if ConvexHull is None or not annots.areas:
        return _handlers_annotations

    for convex_area in annots.areas:
        int_points = np.array(
            [annots.get_point_by_name(key) for key in convex_area.point_names]
        )
        try:
            assert len(int_points) >= 3
            hull = ConvexHull(int_points)
        except (AssertionError, QhullError):  # pragma: no cover
            logging.error(f"QhullError with points: {int_points}")
            continue

        for simplex in hull.simplices:
            _handlers_annotations.append(
                ax.plot(
                    int_points[simplex, 0],
                    int_points[simplex, 1],
                    **convex_area.line_style,
                )
            )
        _handlers_annotations.append(
            ax.fill(
                int_points[hull.vertices, 0],
                int_points[hull.vertices, 1],
                **convex_area.fill_style,
            )
        )

    return _handlers_annotations



from pydantic import Extra, Field, root_validator, validator

from psychrochart.models.base import BaseConfig
from psychrochart.models.validators import parse_color, reduce_field_abrs


class CurveStyle(BaseConfig):
    """Container for matplotlib styling of psychro curve."""

    color: list[float] = Field(default=[0.2, 0.2, 0.2])
    linewidth: float = Field(default=1)
    linestyle: str = Field(default="-")

    class Config:
        extra = Extra.allow

    @validator("color", pre=True, always=True)
    def _color_arr(cls, v, values):
        return parse_color(v)

    @root_validator(pre=True)
    def _remove_aliases(cls, values):
        return reduce_field_abrs(values)


class LabelStyle(BaseConfig):
    """Container for matplotlib styling of labels."""

    color: list[float] = Field(default=[0.2, 0.2, 0.2])
    fontsize: int | float = Field(default=9)

    class Config:
        extra = Extra.allow

    @validator("color", pre=True, always=True)
    def _color_arr(cls, v, values):
        return parse_color(v)


class TickStyle(BaseConfig):
    """Container for matplotlib tick axes styling."""

    direction: str = Field(default="out")
    color: list[float] = Field(default=[0.2, 0.2, 0.2])

    class Config:
        extra = Extra.allow

    @validator("color", pre=True, always=True)
    def _color_arr(cls, v, values):
        return parse_color(v)


class ZoneStyle(BaseConfig):
    """Container for matplotlib patch styling of zones in psychrochart."""

    edgecolor: list[float]
    facecolor: list[float]
    linewidth: float = Field(default=2)
    linestyle: str = Field(default="--")

    class Config:
        extra = Extra.allow

    @validator("edgecolor", "facecolor", pre=True, always=True)
    def _color_arr(cls, v, values):
        return parse_color(v)

    @root_validator(pre=True)
    def _remove_aliases(cls, values):
        return reduce_field_abrs(values)



from pathlib import Path
from typing import Any, Iterable, Type, TypeVar

import numpy as np
from pydantic import BaseModel, parse_obj_as

from psychrochart.chartdata import gen_points_in_constant_relative_humidity
from psychrochart.models.annots import (
    ChartArea,
    ChartLine,
    ChartPoint,
    ChartSeries,
    ChartZones,
    ConvexGroupTuple,
    DEFAULT_ZONES,
)
from psychrochart.models.config import ChartConfig

PATH_STYLES = Path(__file__).absolute().parents[1] / "chart_styles"
DEFAULT_CHART_CONFIG_FILE = str(PATH_STYLES / "default_chart_config.json")
ASHRAE_CHART_CONFIG_FILE = str(PATH_STYLES / "ashrae_chart_style.json")
ASHRAE_IP_CHART_CONFIG_FILE = str(PATH_STYLES / "ashrae_ip_chart_style.json")
INTERIOR_CHART_CONFIG_FILE = str(PATH_STYLES / "interior_chart_style.json")
MINIMAL_CHART_CONFIG_FILE = str(PATH_STYLES / "minimal_chart_style.json")
DEFAULT_ZONES_FILE = str(PATH_STYLES / "default_comfort_zones.json")

STYLES = {
    "ashrae": ASHRAE_CHART_CONFIG_FILE,
    "ashrae_ip": ASHRAE_IP_CHART_CONFIG_FILE,
    "default": DEFAULT_CHART_CONFIG_FILE,
    "interior": INTERIOR_CHART_CONFIG_FILE,
    "minimal": MINIMAL_CHART_CONFIG_FILE,
}

T = TypeVar("T", bound=BaseModel)


def obj_loader(
    data_cls: Type[T],
    data: T | dict[str, Any] | str | None = None,
    default_obj: T | None = None,
) -> T:
    """Load pydantic model from disk, or raw, with optional defaults."""
    if data is None:
        return default_obj if default_obj is not None else data_cls()
    if isinstance(data, data_cls):
        return data
    if isinstance(data, str) and data in STYLES:
        return data_cls.parse_file(STYLES[data])
    if isinstance(data, str):
        return data_cls.parse_file(data)
    return data_cls(**data)


def load_config(
    config: ChartConfig | dict[str, Any] | str | None = None
) -> ChartConfig:
    """Load the plot params for the psychrometric chart."""
    return obj_loader(ChartConfig, config)


def load_zones(
    zones: ChartZones | dict[str, Any] | str | None = None
) -> ChartZones:  # pragma: no cover
    """Load a zones JSON file to overlay in the psychrometric chart."""
    return obj_loader(ChartZones, zones, default_obj=DEFAULT_ZONES)


def load_points_dbt_rh(
    points: dict[str, Any],
    pressure: float,
    scatter_style: dict[str, Any] | None = None,
) -> tuple[dict[str, ChartPoint], dict[str, ChartSeries]]:
    """Load chart single-point + point-array annots with multiple syntaxes."""
    points_plot: dict[str, ChartPoint] = {}
    series_plot: dict[str, ChartSeries] = {}
    default_style = {
        "marker": "o",
        "markersize": 10,
        "color": [1.0, 0.8, 0.1, 0.8],
        "linewidth": 0,
    }
    if scatter_style is not None:
        default_style = scatter_style
    for point_name, point_data in points.items():
        label: str | None = None
        plot_params = default_style.copy()
        if isinstance(point_data, dict) and isinstance(
            point_data["xy"][0], Iterable
        ):
            # arrays with 'xy' notation inside dict
            temperatures, rh_data = point_data["xy"]
            plot_params.update(point_data.get("style", {}))
            label = point_data.get("label")

            w_g_ka = gen_points_in_constant_relative_humidity(
                temperatures, rh_data, pressure
            )
            series_plot[point_name] = ChartSeries(
                x_data=temperatures,
                y_data=w_g_ka,
                style=plot_params,
                label=label,
            )
        elif isinstance(point_data, dict):
            # single points with 'xy' notation inside dict
            x_data, y_data = point_data["xy"]
            plot_params.update(point_data.get("style", {}))
            label = point_data.get("label")

            w_g_ka = gen_points_in_constant_relative_humidity(
                [x_data], y_data, pressure
            )
            points_plot[point_name] = ChartPoint(
                xy=(x_data, w_g_ka[0]),
                style=plot_params,
                label=label,
            )
        elif isinstance(point_data, list) or isinstance(point_data, tuple):
            # simple labeled data
            x_data = point_data[0]
            y_data = point_data[1]
            if isinstance(x_data, Iterable):
                # simple labeled array
                temperatures = np.array(x_data)
                w_g_ka = gen_points_in_constant_relative_humidity(
                    temperatures, y_data, pressure
                )
                series_plot[point_name] = ChartSeries(
                    x_data=temperatures,
                    y_data=w_g_ka,
                    style=plot_params,
                    label=label,
                )
            else:
                # simple labeled point
                w_g_ka = gen_points_in_constant_relative_humidity(
                    [x_data], y_data, pressure
                )
                points_plot[point_name] = ChartPoint(
                    xy=(x_data, w_g_ka[0]),
                    style=plot_params,
                    label=label,
                )

    return points_plot, series_plot


def load_extra_annots(
    connectors: list[dict[str, Any]] | None = None,
    convex_groups: list[dict[str, Any]] | list[ConvexGroupTuple] | None = None,
) -> tuple[list[ChartLine], list[ChartArea]]:
    """Load chart lines + convex groups annotations."""
    data_connectors = parse_obj_as(list[ChartLine], connectors or [])
    data_areas = []
    if convex_groups and isinstance(convex_groups[0], dict):
        data_areas = parse_obj_as(list[ChartArea], convex_groups)
    elif convex_groups and isinstance(convex_groups[0], tuple):
        data_areas = [
            ChartArea.from_tuple(it) for it in convex_groups  # type: ignore
        ]
    return data_connectors, data_areas



from typing import AbstractSet, Any, Mapping

import numpy as np
from pydantic import BaseModel, Field, root_validator

from psychrochart.models.styles import CurveStyle, ZoneStyle
from psychrochart.models.validators import parse_curve_arrays


class PsychroCurve(BaseModel):
    """Pydantic model to store a psychrometric curve for plotting."""

    x_data: np.ndarray
    y_data: np.ndarray
    style: ZoneStyle | CurveStyle
    type_curve: str | None = None
    label: str | None = None
    label_loc: float = 0.75
    internal_value: float | None = None

    class Config:
        arbitrary_types_allowed = True
        json_encoders = {np.ndarray: lambda x: x.tolist()}

    @root_validator(pre=True)
    def _parse_curve_data(cls, values):
        if (
            values.get("label") is None
            and values.get("internal_value") is None
        ):
            raise ValueError(
                "PsychroCurve should have a 'label' or an 'internal_value'"
            )
        return parse_curve_arrays(values)

    @property
    def curve_id(self) -> str:
        """Get Curve identifier (value or label)."""
        if self.internal_value is not None:
            return f"{self.internal_value:g}"
        assert self.label is not None
        return self.label

    def dict(
        self,
        *,
        include: AbstractSet[int | str]
        | Mapping[int | str, Any]
        | None = None,
        exclude: AbstractSet[int | str]
        | Mapping[int | str, Any]
        | None = None,
        by_alias: bool = False,
        skip_defaults: bool | None = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> dict[str, Any]:
        """Override pydantic.BaseModel.dict() to export arrays as list."""
        plain_dict = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )
        plain_dict["x_data"] = plain_dict["x_data"].tolist()
        plain_dict["y_data"] = plain_dict["y_data"].tolist()
        return plain_dict

    def __repr__(self) -> str:
        """Object string representation."""
        name = (
            "PsychroZone"
            if isinstance(self.style, ZoneStyle)
            else "PsychroCurve"
        )
        extra = f" (label: {self.label})" if self.label else ""
        return f"<{name} {len(self.x_data)} values{extra}>"


class PsychroCurves(BaseModel):
    """Pydantic model to store a list of psychrometric curves for plotting."""

    curves: list[PsychroCurve] = Field(min_items=1)
    family_label: str | None = None

    def __repr__(self) -> str:
        """Object string representation."""
        extra = f" (label: {self.family_label})" if self.family_label else ""
        return f"<{len(self.curves)} PsychroCurves{extra}>"


class PsychroChartModel(BaseModel):
    """Pydantic model to store all psychrometric curves for PsychroChart."""

    unit_system_si: bool
    altitude_m: int
    pressure: float

    saturation: PsychroCurves
    constant_dry_temp_data: PsychroCurves | None = None
    constant_humidity_data: PsychroCurves | None = None
    constant_rh_data: PsychroCurves | None = None
    constant_h_data: PsychroCurves | None = None
    constant_v_data: PsychroCurves | None = None
    constant_wbt_data: PsychroCurves | None = None
    zones: list[PsychroCurves] = Field(default_factory=list)



from pydantic import BaseModel


class BaseConfig(BaseModel):
    """
    Pydantic model as common base for configuration models.

    To control chart config changes to trigger model-data regenereration.
    """

    _has_changed: bool = False

    class Config:
        allow_mutation = True
        validate_assignment = True
        underscore_attrs_are_private = True

    def __setattr__(self, name, value):
        """Override setattr to control mutability flag '_has_changed'."""
        super().__setattr__(name, value)
        if name != "_has_changed":
            self._has_changed = True

    @property
    def has_changed(self) -> bool:
        """
        Test if the object has been modified since creation, including childs.
        """
        return self._has_changed or any(
            getattr(field, "has_changed", False)
            for field in self.__dict__.values()
        )

    def commit_changes(self):
        """Reset mutation flag 'has_changed' for the object and its childs."""
        if self.has_changed:
            self._has_changed = False
            for field in self.__dict__.values():
                if hasattr(field, "commit_changes") and field.has_changed:
                    field.commit_changes()



"""A library to make psychrometric charts and overlay information in them."""
import matplotlib.pyplot as plt

from psychrochart.chart import PsychroChart


def main():
    """CLI entry point to show the default psychrometric chart."""
    PsychroChart.create().plot(ax=plt.gca())


if __name__ == "__main__":
    main()
    plt.show()



from pydantic import Extra, Field

from psychrochart.models.styles import (
    BaseConfig,
    CurveStyle,
    LabelStyle,
    TickStyle,
)

_DEFAULT_RANGE_VOL_M3_KG = (0.78, 0.98)
_DEFAULT_RANGE_ENTALPHY = (5, 155)
_DEFAULT_RANGE_WET_TEMP = (-10, 40)
_DEFAULT_POSITION = [0.025, 0.075, 0.925, 0.875]

_DEFAULT_RH_CURVES = [10, 20, 30, 40, 45, 50, 55, 60, 70, 80, 90]
_DEFAULT_CONSTANT_RH_LABELS = [20, 30, 40, 50, 60, 70, 80]
_DEFAULT_CONSTANT_V_LABELS = [0.8, 0.9, 0.96]
_DEFAULT_CONSTANT_H_LABELS = [5, 25, 50, 75, 100, 125]
_DEFAULT_CONSTANT_WET_TEMP_LABELS = [0, 5, 10, 15, 20, 25, 30, 35]

_DEFAULT_X_AXIS = CurveStyle(color=[0.2, 0.2, 0.2], linewidth=2)
_DEFAULT_Y_AXIS = CurveStyle(color=[0.3, 0.3, 0.3], linewidth=2)
_DEFAULT_X_AXIS_LABELS = LabelStyle(color=[0.2, 0.2, 0.2], fontsize=9)
_DEFAULT_Y_AXIS_LABELS = LabelStyle(color=[0.3, 0.3, 0.3], fontsize=8)
_DEFAULT_X_AXIS_TICKS = TickStyle(color=[0.2, 0.2, 0.2], direction="out")
_DEFAULT_Y_AXIS_TICKS = TickStyle(color=[0.3, 0.3, 0.3], direction="out")
_DEFAULT_STYLE_SAT = CurveStyle(color=[0.855, 0.004, 0.278], linewidth=3)
_DEFAULT_STYLE_RH = CurveStyle(color=[0.0, 0.498, 1.0], linestyle="-.")
_DEFAULT_STYLE_V = CurveStyle(color=[0.0, 0.502, 0.337], linestyle="--")
_DEFAULT_STYLE_H = CurveStyle(color=[0.251, 0.0, 0.502], linestyle="--")
_DEFAULT_STYLE_WET_TEMP = CurveStyle(color=[0.498, 0.875, 1.0], linestyle="-.")
_DEFAULT_STYLE_DRY_TEMP = CurveStyle(
    color=[0.855, 0.145, 0.114], linewidth=0.75, linestyle=":"
)
_DEFAULT_STYLE_HUMID = CurveStyle(
    color=[0.0, 0.125, 0.376], linewidth=0.75, linestyle=":"
)


class ChartFigure(BaseConfig):
    """Psychrochart settings for matplotlib Figure."""

    figsize: tuple[float, float] = Field(default=(16, 9))
    dpi: int = Field(default=150)
    fontsize: int = Field(default=10)
    title: str | None = Field(default="Psychrometric Chart")
    x_label: str | None = Field(default="Dry-bulb temperature, $°C$")
    y_label: str | None = Field(default="Humidity ratio $w, g_w / kg_{da}$")

    x_axis: CurveStyle = Field(default=_DEFAULT_X_AXIS)
    x_axis_labels: LabelStyle = Field(default=_DEFAULT_X_AXIS_LABELS)
    y_axis: CurveStyle = Field(default=_DEFAULT_Y_AXIS)
    y_axis_labels: LabelStyle = Field(default=_DEFAULT_Y_AXIS_LABELS)
    x_axis_ticks: TickStyle | None = Field(default=_DEFAULT_X_AXIS_TICKS)
    y_axis_ticks: TickStyle | None = Field(default=_DEFAULT_Y_AXIS_TICKS)

    partial_axis: bool = Field(default=True)
    position: list[float] = Field(
        default_factory=lambda: list(_DEFAULT_POSITION)
    )


class ChartLimits(BaseConfig):
    """Psychrochart temperature + humidity + pressure limits."""

    range_temp_c: tuple[float, float] = Field(default=(0, 50))
    range_humidity_g_kg: tuple[float, float] = Field(default=(0, 40))
    altitude_m: int = Field(default=0)
    pressure_kpa: float | None = Field(default=None)
    step_temp: float = Field(default=1)


class ChartParams(BaseConfig):
    """Psychrochart plotting parameters."""

    with_constant_rh: bool = Field(default=True)
    constant_rh_label: str | None = Field(default="Constant relative humidity")
    constant_rh_curves: list[int] = Field(default=_DEFAULT_RH_CURVES)
    constant_rh_labels: list[int] = Field(default=_DEFAULT_CONSTANT_RH_LABELS)
    constant_rh_labels_loc: float = Field(default=0.85)

    with_constant_v: bool = Field(default=True)
    constant_v_label: str | None = Field(default="Constant specific volume")
    constant_v_step: float = Field(default=0.02)
    range_vol_m3_kg: tuple[float, float] = Field(
        default=_DEFAULT_RANGE_VOL_M3_KG
    )
    constant_v_labels: list[float] = Field(default=_DEFAULT_CONSTANT_V_LABELS)
    constant_v_labels_loc: float = Field(default=1.0)

    with_constant_h: bool = Field(default=True)
    constant_h_label: str | None = Field(default="Constant enthalpy")
    constant_h_step: float = Field(default=5)
    range_h: tuple[float, float] = Field(default=_DEFAULT_RANGE_ENTALPHY)
    constant_h_labels: list[float] = Field(default=_DEFAULT_CONSTANT_H_LABELS)
    constant_h_labels_loc: float = Field(default=1.0)

    with_constant_wet_temp: bool = Field(default=True)
    constant_wet_temp_label: str | None = Field(
        default="Constant wet bulb temperature"
    )
    constant_wet_temp_step: float = Field(default=5)
    range_wet_temp: tuple[float, float] = Field(
        default=_DEFAULT_RANGE_WET_TEMP
    )
    constant_wet_temp_labels: list[float] = Field(
        default=_DEFAULT_CONSTANT_WET_TEMP_LABELS
    )
    constant_wet_temp_labels_loc: float = Field(default=0.05)

    with_constant_dry_temp: bool = Field(default=True)
    constant_temp_label: str | None = Field(default="Dry bulb temperature")
    constant_temp_step: float = Field(default=1)
    constant_temp_label_step: float = Field(default=5)
    constant_temp_label_include_limits: bool = Field(default=True)
    with_constant_humidity: bool = Field(default=True)
    constant_humid_label: str | None = Field(default="Absolute humidity")
    constant_humid_step: float = Field(default=1)
    constant_humid_label_step: float = Field(default=2)
    constant_humid_label_include_limits: bool = Field(default=True)

    with_zones: bool = Field(default=True)


class ChartConfig(BaseConfig):
    """
    Psychrochart configuration model.

    Includes:
    * matplotlib figure settings and styling
    * chart limits
    * plotting parameters
    * plot styling for each kind of psychrometric curve
    """

    figure: ChartFigure = Field(default_factory=ChartFigure)
    limits: ChartLimits = Field(default_factory=ChartLimits)

    saturation: CurveStyle = Field(default=_DEFAULT_STYLE_SAT)
    constant_rh: CurveStyle = Field(default=_DEFAULT_STYLE_RH)
    constant_v: CurveStyle = Field(default=_DEFAULT_STYLE_V)
    constant_h: CurveStyle = Field(default=_DEFAULT_STYLE_H)
    constant_wet_temp: CurveStyle = Field(default=_DEFAULT_STYLE_WET_TEMP)
    constant_dry_temp: CurveStyle = Field(default=_DEFAULT_STYLE_DRY_TEMP)
    constant_humidity: CurveStyle = Field(default=_DEFAULT_STYLE_HUMID)

    chart_params: ChartParams = Field(default_factory=ChartParams)

    class Config:
        extra = Extra.ignore

    @property
    def dbt_min(self) -> float:
        """Left limit for chart."""
        return self.limits.range_temp_c[0]

    @property
    def dbt_max(self) -> float:
        """Right limit for chart."""
        return self.limits.range_temp_c[1]

    @property
    def w_min(self) -> float:
        """Bottom limit for chart."""
        return self.limits.range_humidity_g_kg[0]

    @property
    def w_max(self) -> float:
        """Top limit for chart."""
        return self.limits.range_humidity_g_kg[1]



"""A library to make psychrometric charts and overlay information in them."""
import gc
from io import StringIO
from pathlib import Path
from typing import Any, Iterable, Type

from matplotlib import figure
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.backend_bases import FigureCanvasBase
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.backends.backend_svg import FigureCanvasSVG
from matplotlib.legend import Legend

from psychrochart.chartdata import (
    gen_points_in_constant_relative_humidity,
    make_constant_dry_bulb_v_line,
)
from psychrochart.models.annots import ChartAnnots, ChartZones
from psychrochart.models.config import ChartConfig
from psychrochart.models.curves import PsychroChartModel
from psychrochart.models.parsers import (
    ConvexGroupTuple,
    load_extra_annots,
    load_points_dbt_rh,
    obj_loader,
)
from psychrochart.models.styles import CurveStyle
from psychrochart.plot_logic import (
    add_label_to_curve,
    apply_axis_styling,
    plot_annots_dbt_rh,
    plot_chart,
    plot_curve,
)
from psychrochart.process_logic import (
    append_zones_to_chart,
    generate_psychrochart,
    update_psychrochart_data,
)
from psychrochart.util import mod_color


def _select_fig_canvas(
    path_dest: Any, canvas_cls: Type[FigureCanvasBase] | None = None
) -> Type[FigureCanvasBase]:
    if (
        canvas_cls is None
        and isinstance(path_dest, (str, Path))
        and str(path_dest).endswith(".svg")
    ):
        canvas_cls = FigureCanvasSVG
    elif canvas_cls is None:
        canvas_cls = FigureCanvasAgg
    return canvas_cls


class PsychroChart(PsychroChartModel):
    """Psychrometric chart object handler."""

    config: ChartConfig
    _fig: figure.Figure | None = None
    _axes: Axes | None = None
    _legend: Legend | None = None
    _handlers_annotations: list[Artist | list[Artist]]

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True

    def _init_private_attributes(self) -> None:
        self._handlers_annotations = []
        super()._init_private_attributes()

    @classmethod
    def create(
        cls,
        config: ChartConfig | dict[str, Any] | str | None = None,
        *,
        extra_zones: ChartZones | dict[str, Any] | str | None = None,
        use_unit_system_si: bool = True,
    ):
        chart_config = obj_loader(ChartConfig, config)
        chart_config.commit_changes()
        chart_data = generate_psychrochart(
            chart_config, extra_zones, use_unit_system_si
        )
        chart = cls(config=chart_config, **chart_data.dict())
        return chart

    def __repr__(self) -> str:
        """Return a string representation of the PsychroChart object."""
        return (
            f"<PsychroChart [{self.config.dbt_min:g}"
            f"->{self.config.dbt_max:g} °C, "
            f"{self.config.w_min:g}"
            f"->{self.config.w_max:g} gr/kg_da]>"
        )

    def _ensure_updated_data(self):
        if self.config.has_changed:
            update_psychrochart_data(self, self.config)

    @property
    def axes(self) -> Axes:
        """Return the Axes object plotting the chart if necessary."""
        self._ensure_updated_data()
        if self._axes is None:
            self.plot()
        assert isinstance(self._axes, Axes)
        return self._axes

    def append_zones(
        self, zones: ChartZones | dict[str, Any] | str | None = None
    ) -> None:
        """Append zones as patches to the psychrometric chart."""
        append_zones_to_chart(self.config, self, zones)

    def plot_points_dbt_rh(
        self,
        points: dict[str, Any],
        connectors: list[dict[str, Any]] | None = None,
        convex_groups: list[dict[str, Any]]
        | list[ConvexGroupTuple]
        | None = None,
        scatter_style: dict[str, Any] | None = None,
    ) -> ChartAnnots:
        """Append individual points, connectors and groups to the plot.

        * Pass a specific style dict to do a scatter plot:
            `scatter_style={'s': 5, 'alpha': .1, 'color': 'darkorange'}`

        * if you are plotting series of points, pass them as numpy arrays:
            `points={'points_series_name': (temp_array, humid_array)}`

        - The syntax to add points is:
        ```
        points = {
            'point_1_name': {
                'label': 'label_for_legend',
                'style': {'color': [0.855, 0.004, 0.278, 0.8],
                          'marker': 'X', 'markersize': 15},
                'xy': (31.06, 32.9)},
            'point_2_name': {
                'label': 'label_for_legend',
                'style': {'color': [0.573, 0.106, 0.318, 0.5],
                          'marker': 'x',
                          'markersize': 10},
                'xy': (29.42, 52.34)},
                # ...
        }
        # Or, using the default style:
        points = {
            'point_1_name': (31.06, 32.9),
            'point_2_name': (29.42, 52.34),
            # ...
        }
        ```

        - The syntax to add connectors between pairs of given points is:
        ```
        connectors = [
            {'start': 'point_1_name',
             'end': 'point_2_name',
             'style': {'color': [0.573, 0.106, 0.318, 0.7],
                       "linewidth": 2, "linestyle": "-."}},
            {'start': 'point_2_name',
             'end': 'point_3_name',
             'style': {'color': [0.855, 0.145, 0.114, 0.8],
                       "linewidth": 2, "linestyle": ":"}},
            # ...
        ]
        ```

        - The syntax to add groups of given points (with more than 3 points)
         to plot a styled convex hull area is:
        ```
        interior_zones = [
            # Zone 1:
            ([point_1_name, point_2_name, point_3_name, ...],  # list of points
             {"color": 'darkgreen', "lw": 0, ...},             # line style
             {"color": 'darkgreen', "lw": 0, ...}),            # filling style

            # Zone 2:
            ([point_7_name, point_8_name, point_9_name, ...],  # list of points
             {"color": 'darkorange', "lw": 0, ...},            # line style
             {"color": 'darkorange', "lw": 0, ...}),           # filling style

            # ...
        ]
        ```
        """
        data_points, data_series = load_points_dbt_rh(
            points, self.pressure, scatter_style
        )
        data_lines, data_areas = load_extra_annots(connectors, convex_groups)
        annots = ChartAnnots(
            points=data_points,
            series=data_series,
            connectors=data_lines,
            areas=data_areas,
            use_scatter=scatter_style is not None,
        )
        self._handlers_annotations.extend(
            plot_annots_dbt_rh(self.axes, annots)
        )
        return annots

    def plot_arrows_dbt_rh(
        self, points_pairs: dict[str, Any]
    ) -> dict[str, Any]:
        """Append individual points to the plot."""
        points_plot = {}
        default_style = {
            "linewidth": 0,
            "color": [1.0, 0.8, 0.1, 0.8],
            "arrowstyle": "wedge",
        }
        for key, pair_point in points_pairs.items():
            plot_params = default_style.copy()
            if isinstance(pair_point, dict):
                if "style" in pair_point and "color" in pair_point["style"]:
                    plot_params["color"] = mod_color(
                        pair_point["style"]["color"], 0.6
                    )  # set alpha
                point1, point2 = pair_point["xy"]
            else:
                point1, point2 = pair_point
            temp1 = point1[0]
            temp2 = point2[0]
            w_g_ka1 = gen_points_in_constant_relative_humidity(
                [temp1], point1[1], self.pressure
            )[0]
            w_g_ka2 = gen_points_in_constant_relative_humidity(
                [temp2], point2[1], self.pressure
            )[0]

            self._handlers_annotations.append(
                self.axes.annotate(
                    "",
                    (temp2, w_g_ka2),
                    xytext=(temp1, w_g_ka1),
                    arrowprops=plot_params,
                )
            )

            points_plot[key] = (temp1, w_g_ka1), (temp2, w_g_ka2), plot_params

        return points_plot

    def plot_vertical_dry_bulb_temp_line(
        self,
        temp: float,
        style: CurveStyle | dict[str, Any] | None = None,
        label: str | None = None,
        reverse: bool = False,
        **label_params,
    ) -> None:
        """Append a vertical line from w_min to w_sat."""
        style_curve = obj_loader(
            CurveStyle, style, default_obj=self.config.constant_dry_temp
        )
        curve = make_constant_dry_bulb_v_line(
            self.config.w_min,
            temp,
            self.pressure,
            style=style_curve,
            type_curve="constant-dbt",
            reverse=reverse,
        )

        if plot_curve(curve, self.axes) and label is not None:
            add_label_to_curve(curve, self.axes, label, **label_params)

    def plot_legend(
        self,
        loc: str = "upper left",
        markerscale: float = 0.9,
        frameon: bool = True,
        fancybox: bool = True,
        edgecolor: str | Iterable = "darkgrey",
        fontsize: float = 15.0,
        labelspacing: float = 1.5,
        **params,
    ) -> None:
        """Append a legend to the psychrochart plot."""
        self._legend = self.axes.legend(
            loc=loc,
            markerscale=markerscale,
            frameon=frameon,
            edgecolor=edgecolor,
            fontsize=fontsize,
            fancybox=fancybox,
            labelspacing=labelspacing,
            **params,
        )

    def plot(self, ax: Axes | None = None) -> Axes:
        """Plot the psychrochart and return the matplotlib Axes instance."""
        self._ensure_updated_data()
        if ax is not None:
            self._fig = ax.get_figure()
        else:
            self._fig = figure.Figure(
                figsize=self.config.figure.figsize,
                dpi=self.config.figure.dpi,
                frameon=False,
            )
            ax = self._fig.add_subplot(position=self.config.figure.position)
        self._axes = ax
        apply_axis_styling(self.config, self._axes)
        return plot_chart(self, self._axes)

    def remove_annotations(self) -> None:
        """Remove the annotations made in the chart to reuse it."""
        for line in self._handlers_annotations:
            if isinstance(line, list):
                line[0].remove()
            else:
                line.remove()
        self._handlers_annotations = []

    def remove_legend(self) -> None:
        """Remove the legend of the chart."""
        if self._legend is not None:
            self._legend.remove()
            self._legend = None

    def save(
        self,
        path_dest: Any,
        canvas_cls: Type[FigureCanvasBase] | None = None,
        **params,
    ) -> None:
        """Write the chart to disk."""
        # ensure destination path if folder does not exist
        if (
            isinstance(path_dest, (str, Path))
            and not Path(path_dest).parent.exists()
        ):
            Path(path_dest).parent.mkdir(parents=True)
        if self._axes is None or self.config.has_changed:
            self.plot()
        assert self._fig is not None
        canvas_use = _select_fig_canvas(path_dest, canvas_cls)
        canvas_use(self._fig).print_figure(path_dest, **params)
        gc.collect()

    def make_svg(self, **params) -> str:
        """Generate chart as SVG and return as text."""
        svg_io = StringIO()
        self.save(svg_io, canvas_cls=FigureCanvasSVG, **params)
        svg_io.seek(0)
        return svg_io.read()

    def close_fig(self) -> None:
        """Close the figure plot."""
        assert self._axes is not None
        assert self._fig is not None
        self.remove_annotations()
        self.remove_legend()
        self._axes.remove()
        self._axes = None
        self._fig.clear()
        self._fig = None
        gc.collect()



import logging
from typing import Any

import numpy as np
import psychrolib as psy
from psychrolib import (
    GetStandardAtmPressure,
    GetUnitSystem,
    IP,
    SetUnitSystem,
    SI,
)
from scipy.interpolate import interp1d

from psychrochart.chartdata import (
    get_rh_max_min_in_limits,
    make_constant_dry_bulb_v_lines,
    make_constant_enthalpy_lines,
    make_constant_humidity_ratio_h_lines,
    make_constant_relative_humidity_lines,
    make_constant_specific_volume_lines,
    make_constant_wet_bulb_temperature_lines,
    make_saturation_line,
    make_zone_curve,
)
from psychrochart.models.annots import ChartZones, DEFAULT_ZONES
from psychrochart.models.config import ChartConfig, ChartLimits
from psychrochart.models.curves import PsychroChartModel, PsychroCurves
from psychrochart.models.parsers import obj_loader

spec_vol_vec = np.vectorize(psy.GetMoistAirVolume)


def set_unit_system(use_unit_system_si: bool = True) -> None:
    """Set unit system for psychrolib."""
    if use_unit_system_si and GetUnitSystem() != SI:
        SetUnitSystem(SI)
        logging.info("[SI units mode] ENABLED")
    elif not use_unit_system_si and GetUnitSystem() != IP:
        SetUnitSystem(IP)
        logging.info("[IP units mode] ENABLED")


def get_pressure_pa(limits: ChartLimits, unit_system_si: bool = True) -> float:
    """Set unit system and process ChartLimits for base pressure for chart."""
    set_unit_system(unit_system_si)
    if limits.pressure_kpa is not None:
        return limits.pressure_kpa * 1000.0  # to Pa
    else:
        return GetStandardAtmPressure(limits.altitude_m)


def append_zones_to_chart(
    config: ChartConfig,
    chart: PsychroChartModel,
    zones: ChartZones | dict[str, Any] | str | None = None,
) -> None:
    """Append zones as patches to the psychrometric chart data-container."""
    zones_use = obj_loader(ChartZones, zones, default_obj=DEFAULT_ZONES).zones
    if zones_use:
        curves = PsychroCurves(
            curves=[
                make_zone_curve(zone, config.limits.step_temp, chart.pressure)
                for zone in zones_use
            ]
        )
        chart.zones.append(curves)


def _generate_chart_curves(
    config: ChartConfig, chart: PsychroChartModel, pressure: float
):
    # check chart limits are not fully above the saturation curve!
    assert (chart.saturation.curves[0].y_data > config.w_min).any()
    # check if sat curve cuts x-axis with T > config.dbt_min
    dbt_min_seen: float | None = None
    if chart.saturation.curves[0].y_data[0] < config.w_min:
        temp_sat_interpolator = interp1d(
            chart.saturation.curves[0].y_data,
            chart.saturation.curves[0].x_data,
            assume_sorted=True,
        )
        dbt_min_seen = temp_sat_interpolator(config.w_min)

    # Dry bulb constant lines (vertical):
    if config.chart_params.with_constant_dry_temp:
        step = config.chart_params.constant_temp_step
        temps_vl = np.arange(config.dbt_min, config.dbt_max, step)
        if dbt_min_seen:
            temps_vl = temps_vl[temps_vl > dbt_min_seen]
        chart.constant_dry_temp_data = make_constant_dry_bulb_v_lines(
            config.w_min,
            pressure,
            temps_vl=temps_vl,
            style=config.constant_dry_temp,
            family_label=config.chart_params.constant_temp_label,
        )
    else:
        chart.constant_dry_temp_data = None

    # Absolute humidity constant lines (horizontal):
    if config.chart_params.with_constant_humidity:
        step = config.chart_params.constant_humid_step
        chart.constant_humidity_data = make_constant_humidity_ratio_h_lines(
            config.dbt_max,
            pressure,
            ws_hl=np.arange(
                config.w_min + step,
                config.w_max + step / 10,
                step,
            ),
            style=config.constant_humidity,
            family_label=config.chart_params.constant_humid_label,
        )
    else:
        chart.constant_humidity_data = None

    # Constant relative humidity curves:
    if config.chart_params.with_constant_rh:
        rh_min, rh_max = get_rh_max_min_in_limits(
            dbt_min_seen or config.dbt_min,
            config.dbt_max,
            config.w_min,
            config.w_max,
            pressure,
        )
        rh_values = sorted(
            rh
            for rh in config.chart_params.constant_rh_curves
            if rh_min < rh < rh_max
        )
        start = (
            config.limits.step_temp * (dbt_min_seen // config.limits.step_temp)
            if dbt_min_seen
            else config.dbt_min
        )
        chart.constant_rh_data = make_constant_relative_humidity_lines(
            start,
            config.dbt_max,
            config.limits.step_temp,
            pressure,
            rh_perc_values=rh_values,
            rh_label_values=config.chart_params.constant_rh_labels,
            style=config.constant_rh,
            label_loc=config.chart_params.constant_rh_labels_loc,
            family_label=config.chart_params.constant_rh_label,
        )
    else:
        chart.constant_rh_data = None

    # Constant enthalpy lines:
    if config.chart_params.with_constant_h:
        step = config.chart_params.constant_h_step
        start, end = config.chart_params.range_h
        chart.constant_h_data = make_constant_enthalpy_lines(
            config.w_min,
            pressure,
            enthalpy_values=np.arange(start, end, step),
            h_label_values=config.chart_params.constant_h_labels,
            style=config.constant_h,
            label_loc=config.chart_params.constant_h_labels_loc,
            family_label=config.chart_params.constant_h_label,
            saturation_curve=chart.saturation.curves[0],
            dbt_min_seen=dbt_min_seen,
        )
    else:
        chart.constant_h_data = None

    # Constant specific volume lines:
    if config.chart_params.with_constant_v:
        step = config.chart_params.constant_v_step
        start, end = config.chart_params.range_vol_m3_kg
        chart.constant_v_data = make_constant_specific_volume_lines(
            config.w_min,
            pressure,
            vol_values=np.arange(start, end, step),
            v_label_values=config.chart_params.constant_v_labels,
            style=config.constant_v,
            label_loc=config.chart_params.constant_v_labels_loc,
            family_label=config.chart_params.constant_v_label,
            saturation_curve=chart.saturation.curves[0],
            dbt_min_seen=dbt_min_seen,
        )
    else:
        chart.constant_v_data = None

    # Constant wet bulb temperature lines:
    if config.chart_params.with_constant_wet_temp:
        step = config.chart_params.constant_wet_temp_step
        start, end = config.chart_params.range_wet_temp
        chart.constant_wbt_data = make_constant_wet_bulb_temperature_lines(
            dbt_min_seen or config.dbt_min,
            config.dbt_max,
            config.w_min,
            config.w_max,
            pressure,
            wbt_values=np.arange(start, end, step),
            wbt_label_values=config.chart_params.constant_wet_temp_labels,
            style=config.constant_wet_temp,
            label_loc=config.chart_params.constant_wet_temp_labels_loc,
            family_label=config.chart_params.constant_wet_temp_label,
        )
    else:
        chart.constant_wbt_data = None


def generate_psychrochart(
    config: ChartConfig,
    extra_zones: ChartZones | dict[str, Any] | str | None = None,
    use_unit_system_si: bool = True,
) -> PsychroChartModel:
    """Create the PsychroChart object."""
    # Set unit system for psychrolib and get standard pressure
    pressure = get_pressure_pa(config.limits, use_unit_system_si)

    # base chart with saturation line:
    chart = PsychroChartModel(
        unit_system_si=use_unit_system_si,
        altitude_m=config.limits.altitude_m,
        pressure=pressure,
        saturation=make_saturation_line(
            config.dbt_min,
            config.dbt_max,
            config.limits.step_temp,
            pressure,
            style=config.saturation,
        ),
    )
    _generate_chart_curves(config, chart, pressure)

    if config.chart_params.with_zones:
        append_zones_to_chart(config, chart, extra_zones)

    return chart


def update_psychrochart_data(
    current_chart: PsychroChartModel, config: ChartConfig
) -> None:
    """Update the PsychroChart data with config changes."""
    if config.limits.has_changed:
        current_chart.altitude_m = config.limits.altitude_m
        current_chart.pressure = get_pressure_pa(
            config.limits, current_chart.unit_system_si
        )

    # regen all curves
    current_chart.saturation = make_saturation_line(
        config.dbt_min,
        config.dbt_max,
        config.limits.step_temp,
        current_chart.pressure,
        style=config.saturation,
    )
    _generate_chart_curves(config, current_chart, current_chart.pressure)
    config.commit_changes()
